<?php

namespace Drupal\paragraphs_ai\Plugin\AiAgent;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\DependencyInjection\DependencySerializationTrait;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\ai_agents\Attribute\AiAgent;
use Drupal\ai_agents\Exception\AgentProcessingException;
use Drupal\ai_agents\PluginBase\AiAgentBase;
use Drupal\ai_agents\PluginInterfaces\AiAgentInterface;
use Drupal\paragraphs\Entity\ParagraphsType;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation of the Paragraph Type Agent.
 */
#[AiAgent(
  id: 'paragraph_type_agent',
  label: new TranslatableMarkup('Paragraph Type Agent'),
)]
class ParagraphsTypeAgent extends AiAgentBase {

  use DependencySerializationTrait;

  /**
   * Questions to ask.
   *
   * @var array
   */
  protected $questions = [];

  /**
   * The full result of the task.
   *
   * @var array
   */
  protected $result;

  /**
   * The full data of the initial task.
   *
   * @var array
   */
  protected $data;

  /**
   * Task type.
   *
   * @var string
   */
  protected $taskType;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * {@inheritDoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $parent_instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $parent_instance->entityFieldManager = $container->get('entity_field.manager');
    return $parent_instance;
  }

  /**
   * {@inheritDoc}
   */
  public function getId() {
    return 'paragraph_type_agent';
  }

  /**
   * {@inheritDoc}
   */
  public function agentsNames() {
    return [
      'Paragraph-Type Agent',
    ];
  }

  /**
   * {@inheritDoc}
   */
  public function agentsCapabilities() {
    return [
      'paragraph_type_agent' => [
        'name' => 'Paragraph Type Agent',
        'description' => "This is capable of adding, editing, or removing basic information on Drupal paragraph types and can also answer questions about paragraph types. This is your main agent to generate new paragraph types. It does not add path or field information. Note: This agent can not create or edit paragraph content - only the paragraph types themselves and the settings that belong to them.",
        'inputs' => [
          'free_text' => [
            'name' => 'Prompt',
            'type' => 'string',
            'description' => 'The prompt to add, edit, remove or ask questions about paragraph types.',
            'default_value' => '',
          ],
        ],
        'outputs' => [
          'answers' => [
            'description' => 'The answers to the questions asked about the paragraph type, or the paragraph type that was generated.',
            'type' => 'string',
          ],
        ],
      ],
    ];
  }

  /**
   * {@inheritDoc}
   */
  public function setData($data) {
    $this->data[] = $data;
  }

  /**
   * {@inheritDoc}
   */
  public function isAvailable() {
    // Check if paragraphs module is installed.
    return $this->agentHelper->isModuleEnabled('paragraphs');
  }

  /**
   * {@inheritDoc}
   */
  public function isNotAvailableMessage() {
    return $this->t('You need to enable the paragraphs module to do this.');
  }

  /**
   * {@inheritDoc}
   */
  public function getRetries() {
    return 2;
  }

  /**
   * {@inheritDoc}
   */
  public function getData() {
    return $this->data;
  }

  /**
   * {@inheritDoc}
   */
  public function answerQuestion() {
    $data = $this->agentHelper->runSubAgent('answerQuestion', [
      'The list of paragraph types' => $this->getVerboseParagraphsTypesAsString(),
    ]);

    $answer = "";
    if (isset($data[0]['answer'])) {
      foreach ($data as $dataPoint) {
        $answer .= $dataPoint['answer'] . "\n";
      }
      return $answer;
    }

    return $this->t("Sorry, I got no answers for you.");
  }

  /**
   * {@inheritDoc}
   */
  public function getHelp() {
    $help = $this->t("This agent can figure out paragraph types of a file. Just upload and ask.");
    return $help;
  }

  /**
   * {@inheritDoc}
   */
  public function hasAccess() {
    // Check for permissions.
    if (!$this->currentUser->hasPermission('administer paragraphs types')) {
      return AccessResult::forbidden();
    }
    return parent::hasAccess();
  }

  /**
   * {@inheritDoc}
   */
  public function determineSolvability() {
    parent::determineSolvability();
    $this->taskType = $this->determineTypeOfTask();
    switch ($this->taskType) {
      case 'create':
      case 'blueprint':
        return AiAgentInterface::JOB_SOLVABLE;

      case 'edit':
        return AiAgentInterface::JOB_SOLVABLE;

      case 'delete':
        return AiAgentInterface::JOB_NEEDS_ANSWERS;

      case 'information':
        return AiAgentInterface::JOB_SHOULD_ANSWER_QUESTION;

      case 'fail':
        return AiAgentInterface::JOB_INFORMS;
    }

    return AiAgentInterface::JOB_NOT_SOLVABLE;
  }

  /**
   * {@inheritDoc}
   */
  public function askQuestion() {
    return $this->questions;
  }

  /**
   * {@inheritDoc}
   */
  public function solve() {
    parent::solve();
    switch ($this->data[0]['action']) {
      case 'create':
        $this->createParagraphsType();
        break;

      case 'blueprint':
        return $this->data[0];

      case 'edit':
        $this->editParagraphsType();
        break;

      case 'delete':
        $this->result[] = "The agent does not have permissions to delete node types, but you can ask me for instructions to do it yourself.";
        break;

      default:
        $this->result[] = 'We could not figure out what you wanted to do.';
    }
    return implode("\n", $this->result);
  }

  /**
   * {@inheritDoc}
   */
  public function approveSolution() {
    $this->data[0]['action'] = 'create';
  }

  /**
   * Check so all requirements are there.
   *
   * @return bool
   *   If all requirements are there.
   */
  public function checkRequirements() {
    return TRUE;
  }

  /**
   * Determine if the context is asking a question or wants a audit done.
   *
   * @return string
   *   The context.
   */
  public function determineTypeOfTask() {
    $data = $this->agentHelper->runSubAgent('determineParagraphsTypeTask', [
      'All currently available paragraph types' => $this->getParagraphsTypesAsString(),
    ]);

    if (isset($data[0]['action']) && in_array($data[0]['action'], [
      'create',
      'edit',
      'delete',
      'information',
      'fail',
    ])) {
      // If its edit or delete, we need to know if the paragraphs type exists.
      if (in_array($data[0]['action'], ['edit', 'delete'])) {
        if ($data[0]['action'] == 'delete') {
          $this->questions[] = "You do not have permission to use the delete function, if you feel you need it to answer the user's request, please inform the user that you don't have permission and that the user can ask for help doing it themselves or obtaining permission from an admin.";
        }
        if (!$this->doesParagraphsTypeExist($data[0]['data_name'])) {
          // Check if we can find a similar paragraphs type.
          $similarParagraphsType = $this->findSimilarParagraphsType();
          // Fail completely.
          if (empty($similarParagraphsType)) {
            throw new \Exception('We could not figure out which paragraphs type you meant. Please specify better.');
          }
          // If there is a similar, we ask.
          $this->information = 'We could not find the paragraphs type you are looking for. Do you mean ' . $similarParagraphsType['readable_name'] . ' (' . $similarParagraphsType['data_name'] . ')?';
          $data[0]['action'] = 'fail';
        }
      }
      // Set blueprint.
      if ($data[0]['action'] === 'create' && !$this->createDirectly) {
        $data[0]['action'] = 'blueprint';
      }
      // Inform the user of failure.
      if ($data[0]['action'] === 'fail') {
        $this->information = $data[0]['fail_message'];
      }

      $this->data = $data;
      return $data[0]['action'];
    }
    throw new \Exception('Invalid action in Web Determining task.');
  }

  /**
   * Try to get a similar paragraphs type.
   *
   * @return string|array
   *   The paragraph type.
   */
  public function findSimilarParagraphsType() {
    $paragraphsTypes = $this->getParagraphsTypes();
    $list = "";
    foreach ($paragraphsTypes as $id => $paragraphsType) {
      $list .= $paragraphsType . ' - id: ' . $id . "\n";
    }
    $data = $this->agentHelper->runSubAgent('determineParagraphsType', [
      'Paragraphs Types List' => $list,
    ]);
    return !empty($data[0]['paragraphs_type']) ? [
      'data_name' => $data[0]['paragraphs_type'],
      'readable_name' => $paragraphsTypes[$data[0]['paragraphs_type']],
    ] : '';
  }

  /**
   * Does the paragraphs type exist.
   *
   * @param string $dataName
   *   The data name.
   *
   * @return bool
   *   If the paragraphs type exists.
   */
  public function doesParagraphsTypeExist($dataName) {
    $paragraphsTypes = $this->getParagraphsTypes();
    return in_array($dataName, array_keys($paragraphsTypes));
  }

  /**
   * Delete a paragraphs type.
   *
   * @return \Drupal\paragraphs\Entity\ParagraphsType
   *   The node type.
   */
  public function deleteParagraphsType() {
    $type = ParagraphsType::load($this->data[0]['data_name']);
    if (!$type) {
      throw new \Exception('We could not figure out which paragraphs type you meant. Please specify better or check the list %here.');
    }

    $type->delete();
    return $type;
  }

  /**
   * Edit a paragraphs type.
   */
  public function editParagraphsType() {
    $types = $this->getParagraphsTypes();
    if (!isset($types[$this->data[0]['data_name']])) {
      throw new \Exception('We could not figure out which paragraphs type you meant. Please specify better or check the list %here.');
    }
    $paragraph = ParagraphsType::load($this->data[0]['data_name']);
    // Set the original.
    $this->setOriginalConfigurations($paragraph);

    // On the paragraph.
    foreach ([
      'readable_name' => 'label',
      'description' => 'description',
    ] as $part => $field) {
      if (isset($this->data[0][$part])) {
        $paragraph->set($field, $this->data[0][$part]);
      }
    }

    if ($paragraph->save()) {
      // Check the diff.
      $url = Url::fromRoute('entity.paragraphs_type.edit_form', ['paragraphs_type' => $paragraph->id()]);

      $diff = $this->getDiffOfConfigurations($paragraph);
      $this->result[] = $this->t('The paragraphs type %paragraphs_type (%data_name) has been updated. Its available to check under @link.', [
        '%paragraphs_type' => $this->data[0]['readable_name'],
        '%data_name' => $this->data[0]['data_name'],
        '@link' => $url->toString(),
      ]);
      $this->structuredResultData->setEditedConfig($paragraph, $diff);
    }
    else {
      throw new AgentProcessingException('There was an error updating the paragraphs type.');
    }
  }

  /**
   * Create a paragraphs type.
   */
  public function createParagraphsType() {
    $paragraph = ParagraphsType::create([
      'id' => $this->data[0]['data_name'],
      'label' => $this->data[0]['readable_name'],
      'description' => $this->data[0]['description'],
    ]);
    if ($paragraph->save()) {
      $url = Url::fromRoute('entity.paragraphs_type.edit_form', ['paragraphs_type' => $paragraph->id()]);

      $this->result[] = $this->t('The paragraphs type %paragraphs_type (%data_name) has been created. Its available to check under @link.', [
        '%paragraphs_type' => $this->data[0]['readable_name'],
        '%data_name' => $this->data[0]['data_name'],
        '@link' => $url->toString(),
      ]);
      $this->structuredResultData->setCreatedConfig($paragraph);
    }
    else {
      throw new AgentProcessingException('There was an error creating the paragraphs type.');
    }
  }

  /**
   * Get all node types.
   *
   * @return array
   *   The paragraphs types.
   */
  public function getParagraphsTypes() {
    $types = ParagraphsType::loadMultiple();
    $paragraphsTypes = [];
    foreach ($types as $type) {
      $paragraphsTypes[$type->id()] = $type->label();
    }
    return $paragraphsTypes;
  }

  /**
   * Get all paragraphs types as a string.
   *
   * @return string
   *   The paragraphs types.
   */
  public function getParagraphsTypesAsString() {
    $paragraphsTypes = $this->getParagraphsTypes();
    $list = "";
    foreach ($paragraphsTypes as $dataName => $paragraphsType) {
      $list .= $paragraphsType . ' - dataname: ' . $dataName . "\n";
    }
    return $list;
  }

  /**
   * Get all paragraphs types as a string with verbose information.
   *
   * @return string
   *   The paragraphs types.
   */
  public function getVerboseParagraphsTypesAsString() {
    $paragraphsTypes = $this->getParagraphsTypes();
    $list = "";
    foreach ($paragraphsTypes as $dataName => $paragraphType) {
      // Load the paragraph type.
      $type = ParagraphsType::load($dataName);
      // Show all the configurations.
      $list .= $paragraphType . ' - dataname: ' . $dataName . "\n";
      $list .= 'Description: ' . $type->getDescription() . "\n";
      $list .= "\n";
    }
    return $list;
  }

}
